import numpy as np
from scipy.ndimage import distance_transform_edt


def add_soft_edge(mask: np.ndarray, width: int) -> np.ndarray:
    distance_from_edge = distance_transform_edt(mask == 0)
    boundary_pixels = (distance_from_edge <= width) & (distance_from_edge != 0)
    normalised_distance_from_edge = (distance_from_edge[boundary_pixels] / width) * np.pi
    output = np.copy(mask)
    output[boundary_pixels] = (0.5 * np.cos(normalised_distance_from_edge) + 0.5)
    return output



from ttmask import cone, cube, cuboid, cylinder, ellipsoid, map2mask, sphere, tube
import numpy as np

def test_cone():
    mask = cone(100, 50, 50, 0, 1, "standard")
    assert mask.shape == (100, 100, 100)
    assert mask.sum() > np.pi * 24**2 * (50 / 3) # Volume of cone
    assert mask.sum() < np.pi * 25**2 * 50 # Volume of cylinder

def test_cube():
    mask = cube(100, 50, 0, 0, 1, "standard")
    assert mask.shape == (100, 100, 100)
    # Test against volume of cube +- center and subpixel issues
    assert mask.sum() > 48**3
    assert mask.sum() < 52**3

def test_cuboid():
    mask = cuboid(100, (50,40,30), 0, 0, 1, "standard")
    assert mask.shape == (100, 100, 100)
    # Test against volume of cuboid +- center and subpixel issues
    assert mask.sum() > 48 * 38 * 28
    assert mask.sum() < 52 * 42 * 32

#def test_curved_surface():
#    mask = curved_surface(100, 50, 50, 0, 1)
#    assert mask.shape == (100, 100, 100)
#    assert mask.sum() > 2 * np.pi * 25**2 # Area of cylinder
#    assert mask.sum() < 2 * np.pi * 25 * 50 # Area of cylinder

def test_cylinder():
    mask = cylinder(100, 50, 50, 0, 0, 1, "standard")
    assert mask.shape == (100, 100, 100)
    assert mask.sum() > np.pi * 25**2 * 48 # Volume of cylinder
    assert mask.sum() < np.pi * 25**2 * 51 # Volume of cylinder

def test_ellipsoid():
    mask = ellipsoid(100, (50,40,30), 0, 0, 1, "standard")
    assert mask.shape == (100, 100, 100)
    # Test against volume of ellipsoid +- center and subpixel issues
    assert mask.sum() > 24 * 19 * 14 * 4/3 * np.pi
    assert mask.sum() < 26 * 21 * 16 * 4/3 * np.pi

def test_sphere():
    mask = sphere(100, 50, 0, 0, 1, "standard")
    assert mask.shape == (100, 100, 100)
    assert mask.sum() > 4/3 * np.pi * 24**3 # Volume of sphere
    assert mask.sum() < 4/3 * np.pi * 26**3 # Volume of sphere

def test_tube():
    mask = tube(100, 50, 50, 0, 0, 1, "standard")
    assert mask.shape == (100, 100, 100)
    assert mask.sum() > np.pi * 24**2 * 48 # Volume of tube
    assert mask.sum() < np.pi * 26**2 * 52 # Volume of tube


from pathlib import Path
import numpy as np
import typer
import mrcfile

from .soft_edge import add_soft_edge
from ._cli import cli
from .box_setup import box_setup

def cube(
    sidelength: int, 
    cube_sidelength: float,
    wall_thickness: float,
    soft_edge_width: int,
    pixel_size: float,
    centering: str
) -> np.ndarray:
     # establish our coordinate system and empty mask
    coordinates_centered, mask = box_setup(sidelength, centering)
    #converting relative coordinates to xyz distances (i.e. not a negative number) :
    xyz_distances = np.abs(coordinates_centered)

    # set up criteria for which pixels are inside the cube and modify values to 1.
    in_cube = np.all(xyz_distances < np.array(cube_sidelength) / (pixel_size * 2), axis=-1)
    mask[in_cube] = 1

    # if requested, criteria set up for pixels within the hollowed area and these values changed to zero
    if wall_thickness != 0:
        within_hollowing = np.all(xyz_distances < ((np.array(cube_sidelength) / (pixel_size * 2)) - wall_thickness),
                                  axis=-1)
        mask[within_hollowing] = 0

    #if requested, a soft edge is added to the mask
    mask = add_soft_edge(mask, soft_edge_width)

    return mask

@cli.command(name='cube')
def cube_cli(
    sidelength: int = typer.Option(...),
    cube_sidelength: float = typer.Option(...),
    wall_thickness: float = typer.Option(0),
    soft_edge_width: int = typer.Option(0),
    pixel_size: float = typer.Option(1),
    output: Path = typer.Option(Path("cube.mrc")),
    centering: float = typer.Option("standard"),
):
    mask = cube(sidelength, cube_sidelength, wall_thickness, soft_edge_width, pixel_size, centering)

    # Save the mask to an MRC file
    with mrcfile.new(output, overwrite=True) as mrc:
        mrc.set_data(mask.astype(np.float32))
        mrc.voxel_size = pixel_size



"""CLI tool for mask creation in cryo-EM/ET"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("ttmask")
except PackageNotFoundError:
    __version__ = "uninstalled"
__author__ = "Miles Graham"
__email__ = "miles.graham@balliol.ox.ac.uk"

from ._cli import cli
from .sphere import sphere
from .cylinder import cylinder
from .cuboid import cuboid
from .cube import cube
from .cone import cone
from .ellipsoid import ellipsoid
from .map2mask import map2mask
from .tube import tube
from .curved_surface import curved_surface




from pathlib import Path

import numpy as np
import typer
import mrcfile

from ._cli import cli

from .soft_edge import add_soft_edge
from .box_setup import box_setup

def cylinder(
    sidelength: int, 
    cylinder_height: float, 
    cylinder_diameter: float, 
    wall_thickness: float, 
    soft_edge_width: int,
    pixel_size: float,
    centering: str
) -> np.ndarray:
    cylinder_radius = cylinder_diameter / 2

    # establish our coordinate system and empty mask
    coordinates_centered, mask = box_setup(sidelength, centering)

    # converting relative coordinates to xyz distances (i.e. not a negative number) :
    xyz_distances = np.abs(coordinates_centered)

    # set up criteria for which pixels are inside the cylinder and modify values to 1.

    xy_distance = np.linalg.norm(xyz_distances[:, :, :, [1, 2]], axis=-1)
    within_xy = xy_distance < (cylinder_radius / pixel_size)
    within_z = xyz_distances[:, :, :, 0] < (cylinder_height / (2 * pixel_size))
    mask[np.logical_and(within_z, within_xy)] = 1

    # if requested, criteria set up for pixels within the hollowed area and these values changed to zero
    if wall_thickness != 0:
        within_z_hollowing = xyz_distances[:, :, :, 0] < (cylinder_height - wall_thickness) / (2 * pixel_size)
        within_xy_hollowing = xy_distance < ((cylinder_radius - wall_thickness) / pixel_size)
        mask[np.logical_and(within_z_hollowing, within_xy_hollowing)] = 0

    # if requested, a soft edge is added to the mask
    mask = add_soft_edge(mask, soft_edge_width)

    return mask


@cli.command(name='cylinder')
def cylinder_cli(
    sidelength: int = typer.Option(...),
    cylinder_height: float = typer.Option(...),
    cylinder_diameter: float = typer.Option(...),
    wall_thickness: float = typer.Option(0),
    soft_edge_width: int = typer.Option(0),
    pixel_size: float = typer.Option(1),
    output: Path = typer.Option(Path("cylinder.mrc")),
    centering: str = typer.Option("standard"),
):
    mask = cylinder(sidelength, cylinder_height, cylinder_diameter, wall_thickness, soft_edge_width, pixel_size, centering)

    # Save the mask to an MRC file
    with mrcfile.new(output, overwrite=True) as mrc:
        mrc.set_data(mask.astype(np.float32))
        mrc.voxel_size = pixel_size


from pathlib import Path

import numpy as np
import typer
import mrcfile
from typing import Tuple
from typing_extensions import Annotated

from .soft_edge import add_soft_edge
from ._cli import cli
from .box_setup import box_setup

def ellipsoid(
    sidelength: int, 
    ellipsoid_dimensions: Tuple[float, float, float],
    wall_thickness: float,
    soft_edge_width: int,
    pixel_size: float,
    centering: str
) -> np.ndarray:
    # establish our coordinate system and empty mask
    coordinates_centered, mask = box_setup(sidelength, centering)
    #converting relative coordinates to xyz distances (i.e. not a negative number) :
    xyz_distances = np.abs(coordinates_centered)

    #extract xyz magnitudes from xyz_distances
    x_magnitude = xyz_distances[:, :, :, 2]
    y_magnitude = xyz_distances[:, :, :, 1]
    z_magnitude = xyz_distances[:, :, :, 0]

    #define desired dimensions in pixels (converting from angstrom)
    z_axis_length = ellipsoid_dimensions[0] / (2 * pixel_size)
    y_axis_length = ellipsoid_dimensions[1] / (2 * pixel_size)
    x_axis_length = ellipsoid_dimensions[2] / (2 * pixel_size)

    # set up criteria for which pixels are inside the ellipsoid and modify values to 1.
    in_ellipsoid = (((x_magnitude) ** 2) / (x_axis_length ** 2)) + ((y_magnitude ** 2) / (y_axis_length ** 2)) + (
        (z_magnitude ** 2) / (z_axis_length ** 2)) <= 1
    mask[in_ellipsoid] = 1

    # if requested, criteria set up for pixels within the hollowed area and these values changed to zero
    if wall_thickness != 0:
        in_hollowing = (((x_magnitude) ** 2) / ((x_axis_length - wall_thickness) ** 2)) + (
                (y_magnitude ** 2) / ((y_axis_length - wall_thickness) ** 2)) + (
                           (z_magnitude ** 2) / ((z_axis_length - wall_thickness) ** 2)) <= 1
        mask[in_hollowing] = 0

    # if requested, a soft edge is added to the mask
    mask = add_soft_edge(mask, soft_edge_width)

    return mask

@cli.command(name='ellipsoid')
def ellipsoid_cli(
    sidelength: int = typer.Option(...),
    ellipsoid_dimensions: Annotated[Tuple[float, float, float], typer.Option()] = (None, None, None),
    wall_thickness: float = typer.Option(0),
    soft_edge_width: int = typer.Option(0),
    pixel_size: float = typer.Option(1),
    output: Path = typer.Option(Path("ellipsoid.mrc")),
    centering: str = typer.Option("standard"),
):
    mask = ellipsoid(sidelength, ellipsoid_dimensions, wall_thickness, soft_edge_width, pixel_size, centering)

    # Save the mask to an MRC file
    with mrcfile.new(output, overwrite=True) as mrc:
        mrc.set_data(mask.astype(np.float32))
        mrc.voxel_size = pixel_size



import mrcfile
import numpy as np
import typer
from pathlib import Path

from ._cli import cli
from .soft_edge import add_soft_edge
from .add_padding import add_padding

def mask_from_map(
    map_data: np.ndarray, 
    binarization_threshold: float,
    padding_width: int,
    soft_edge_width: int
) -> np.ndarray:

    above_threshold = map_data >= binarization_threshold
    below_threshold = map_data < binarization_threshold

    map_data[above_threshold] = 1
    map_data[below_threshold] = 0

    padded_mask = add_padding(map_data, padding_width)
    mask = add_soft_edge(padded_mask, soft_edge_width)

    return mask

@cli.command(name='map2mask')
def map2mask(
    input_map: Path = typer.Option(Path("map.mrc")),
    binarization_threshold: float = typer.Option(...),
    padding_width: int = typer.Option(0),
    soft_edge_width: int = typer.Option(0),
    pixel_size: float = typer.Option(...),
    output_mask: Path = typer.Option(Path("mask.mrc")),
):
    with mrcfile.open(input_map, permissive=True) as mrc:
        data = mrc.data
    mask = mask_from_map(data, binarization_threshold, padding_width, soft_edge_width)

    with mrcfile.new(output_mask, overwrite=True) as mrc:
        mrc.set_data(mask.astype(np.float32))
        mrc.voxel_size = pixel_size


from pathlib import Path
import numpy as np
import einops
import typer
import mrcfile

from ._cli import cli
from .soft_edge import add_soft_edge
from .box_setup import box_setup

def cone(
    sidelength: int, 
    cone_height: float, 
    cone_base_diameter: float, 
    soft_edge_width: int,
    pixel_size: float,
    centering: str
) -> np.ndarray:
    # establish our coordinate system and empty mask
    coordinates_centered, mask = box_setup(sidelength, centering)
    # distances between each pixel and center :
    magnitudes = np.linalg.norm(coordinates_centered, axis=-1)
    magnitudes = einops.rearrange(magnitudes, 'd h w -> d h w 1')

    # Check for zeros in magnitudes and replace them with a small value to avoid Nan warning
    near_zero = 1e-8
    magnitudes = np.where(magnitudes == 0, near_zero, magnitudes)
    normalised = coordinates_centered / magnitudes

    principal_axis = np.array([1, 0, 0])
    dot_product = np.dot(normalised, principal_axis)
    angles_radians = np.arccos(dot_product)
    angles = np.rad2deg(angles_radians)

    z_distance = coordinates_centered[:, :, :, 0]  # (100, 100, 100)

    # Calculate the angle from the tip of the cone to the edge of the base
    cone_base_radius = (cone_base_diameter / 2) / pixel_size
    cone_angle = np.rad2deg(np.arctan(cone_base_radius / cone_height))

    within_cone_height = z_distance < (cone_height / pixel_size)
    within_cone_angle = angles < cone_angle

    # mask[within_cone_height] = 1
    mask[np.logical_and(within_cone_height, within_cone_angle)] = 1

    #   need to adjust the center of the hollow cone otherwise the cone thins towards the apex
    # thickness will need to decrease towards the apex anyway - it's surely not possible / realistic?

    # Shift the mask in the z-axis by cone_height / 2
    z_shift = -int(cone_height / 2)
    mask = np.roll(mask, z_shift, axis=0)
    mask = add_soft_edge(mask, soft_edge_width)

    return mask

@cli.command(name='cone')
def cone_cli(
    sidelength: int = typer.Option(...),
    cone_height: float = typer.Option(...),
    cone_base_diameter: float = typer.Option(...),
    soft_edge_width: int = typer.Option(0),
    pixel_size: float = typer.Option(1),
    output: Path = typer.Option(Path("cone.mrc")),
    centering: str = typer.Option("standard"),
):
    mask = cone(sidelength, cone_height, cone_base_diameter, soft_edge_width, pixel_size, centering)

    # Save the mask to an MRC file
    with mrcfile.new(output, overwrite=True) as mrc:
        mrc.set_data(mask.astype(np.float32))
        mrc.voxel_size = pixel_size



import numpy as np
import einops


def box_setup(sidelength: int, centering: str) -> np.ndarray:
    c = sidelength // 2
    if centering == "visual" and sidelength % 2 == 0:
        center = np.array([c, c, c]) - 0.5
    else:
        center = np.array([c, c, c])
    mask = np.zeros(shape=(sidelength, sidelength, sidelength), dtype=np.float32)

    # 3d coordinates of all voxels
    coordinates = np.indices([sidelength, sidelength, sidelength])
    coordinates = einops.rearrange(coordinates, 'zyx d h w -> d h w zyx')

    #coordinates now expressed relative to center
    coordinates_centered = coordinates - center
    return (coordinates_centered, mask)



from pathlib import Path
import numpy as np
import typer
import mrcfile

from ._cli import cli
from .soft_edge import add_soft_edge
from .box_setup import box_setup

def curved_surface(
    sidelength: int, 
    fit_sphere_diameter: float,
    surface_thickness: float,
    soft_edge_width: int,
    pixel_size: float,
    centering: str
) -> np.ndarray:
    sphere_radius = fit_sphere_diameter / 2

    # establish our coordinate system and empty mask
    coordinates_centered, mask = box_setup(sidelength, centering)
    coordinates_shifted = coordinates_centered - ([0, sphere_radius, 0])


    #determine distances of each pixel to the center
    distance_to_center = np.linalg.norm(coordinates_shifted, axis=-1)


    # set up criteria for which pixels are inside the sphere and modify values to 1.
    inside_sphere = distance_to_center < (sphere_radius / pixel_size)
    mask[inside_sphere] = 1

    # if requested, criteria set up for pixels within the hollowed area and these values changed to zero
    if surface_thickness != 0:
        within_hollowing = distance_to_center < ((sphere_radius - surface_thickness) / pixel_size)
        mask[within_hollowing] = 0

    # if requested, a soft edge is added to the mask
    mask = add_soft_edge(mask, soft_edge_width)

    return mask

@cli.command(name='curved_surface')
def curved_surface_cli(
    sidelength: int = typer.Option(...),
    fit_sphere_diameter: float = typer.Option(...),
    surface_thickness: float = typer.Option(...),
    soft_edge_width: int = typer.Option(0),
    pixel_size: float = typer.Option(1),
    output: Path = typer.Option(Path("curved_surface.mrc")),
    centering: str = typer.Option("standard"),
):
    mask = curved_surface(sidelength, fit_sphere_diameter, surface_thickness, soft_edge_width, pixel_size, centering)

    # Save the mask to an MRC file
    with mrcfile.new(output, overwrite=True) as mrc:
        mrc.set_data(mask.astype(np.float32))
        mrc.voxel_size = pixel_size



from pathlib import Path
import numpy as np
import typer
from typing import Tuple
from typing_extensions import Annotated
import mrcfile

from ._cli import cli
from .soft_edge import add_soft_edge
from .box_setup import box_setup

def cuboid(
    sidelength: int, 
    cuboid_sidelengths: Tuple[float, float, float],
    wall_thickness: float,
    soft_edge_width: int,
    pixel_size: float,
    centering: str
) -> np.ndarray:
    # establish our coordinate system and empty mask
    coordinates_centered, mask = box_setup(sidelength, centering)
    #converting relative coordinates to xyz distances (i.e. not a negative number) :
    xyz_distances = np.abs(coordinates_centered)

    # set up criteria for which pixels are inside the cuboid and modify values to 1.
    inside_cuboid = np.all(xyz_distances < (np.array(cuboid_sidelengths) / (2 * pixel_size)), axis=-1)
    mask[inside_cuboid] = 1

    # if requested, criteria set up for pixels within the hollowed area and these values changed to zero
    if wall_thickness != 0:
        within_hollowing = np.all(xyz_distances < ((np.array(cuboid_sidelengths) / (2 * pixel_size)) - wall_thickness),
                                  axis=-1)
        mask[within_hollowing] = 0

    # if requested, a soft edge is added to the mask
    mask = add_soft_edge(mask, soft_edge_width)
    
    return mask

@cli.command(name='cuboid')
def cuboid_cli(
    sidelength: int = typer.Option(...),
    cuboid_sidelengths: Annotated[Tuple[float, float, float], typer.Option()] = (None, None, None),
    wall_thickness: float = typer.Option(0),
    soft_edge_width: int = typer.Option(0),
    pixel_size: float = typer.Option(1),
    output: Path = typer.Option(Path("cuboid.mrc")),
    centering: str = typer.Option("standard"),
):
    mask = cuboid(sidelength, cuboid_sidelengths, wall_thickness, soft_edge_width, pixel_size, centering)

    # Save the mask to an MRC file
    with mrcfile.new(output, overwrite=True) as mrc:
        mrc.set_data(mask.astype(np.float32))
        mrc.voxel_size = pixel_size
